import boto3
import logging
from typing import Optional
from .datetime_utils import convert_utc_datetime_to_useastern_datetime_string

logging.basicConfig()
logger = logging.getLogger(__name__)


class AwsS3:

    @staticmethod
    def get_buckets() -> list:
        results = []
        try:
            s3 = boto3.resource("s3")
            results = sorted([bucket.name for bucket in s3.buckets.all()])
        except Exception as e:
            logger.error(f"Exception getting S3 bucket list: {e}")
        return results

    @staticmethod
    def get_bucket_keys(bucket_name: str) -> list:
        results = []
        try:
            s3 = boto3.client("s3")
            bucket_keys = s3.list_objects(Bucket=bucket_name)
            if bucket_keys:
                bucket_keys = bucket_keys.get("Contents")
                if bucket_keys:
                    for bucket_key in sorted(bucket_keys, key=lambda item: item["Key"]):
                        results.append({
                            "key": bucket_key["Key"],
                            "size": bucket_key["Size"],
                            "modified": convert_utc_datetime_to_useastern_datetime_string(bucket_key["LastModified"])
                        })

        except Exception as e:
            logger.error(f"Exception getting S3 bucket key list: {e}")
        return results

    SHOW_BUCKET_KEY_CONTENT_MAX_SIZE_BYTES = 50000

    @staticmethod
    def _may_look_at_key_content(bucket, key, size) -> bool:
        if size > AwsS3.SHOW_BUCKET_KEY_CONTENT_MAX_SIZE_BYTES:
            return False
        if key.endswith(".json"):
            return True
        if bucket.endswith("-envs"):
            return True
        return False

    @staticmethod
    def _get_bucket_key_content_size(bucket_name: str, bucket_key_name) -> int:
        try:
            s3 = boto3.client('s3')
            response = s3.head_object(Bucket=bucket_name, Key=bucket_key_name)
            size = response['ContentLength']
            return size
        except Exception as e:
            logger.error(f"Exception getting S3 key content size: {e}")

    @staticmethod
    def get_bucket_key_contents(bucket_name: str, bucket_key_name) -> Optional[list]:
        try:
            size = AwsS3._get_bucket_key_content_size(bucket_name, bucket_key_name)
            if not AwsS3._may_look_at_key_content(bucket_name, bucket_key_name, size):
                return None
            s3 = boto3.resource("s3")
            s3_object = s3.Object(bucket_name, bucket_key_name)
            return s3_object.get()["Body"].read().decode("utf-8")
        except Exception as e:
            logger.error(f"Exception getting S3 key content: {e}")



import boto3
import logging
import time
from typing import Optional
from .cookie_utils import read_cookie
from .envs import Envs
from .jwt_utils import JWT_AUDIENCE_PROPERTY_NAME, jwt_decode, jwt_encode

logging.basicConfig()
logger = logging.getLogger(__name__)


class Auth:

    def __init__(self, auth0_client_id: str, auth0_secret: str, envs: Envs):
        self._auth0_client_id = auth0_client_id
        self._auth0_secret = auth0_secret
        self._envs = envs

    _cached_aws_credentials = {}

    def authorize(self, request: dict, env: Optional[str] = None) -> dict:
        """
        Verifies that the given request is authenticated AND authorized, based on the authtoken
        cookie (a JWT-signed-encoded value) in the given request. If so, returns the verified and
        decoded authtoken as a dictionary. If not, returns a dictionary indicating not authorized
        and/or not authenticated, and containing the basic info from the authtoken.
        """
        try:

            # Read the authtoken cookie.

            authtoken = read_cookie(request, "authtoken")
            if not authtoken:
                return self._create_unauthenticated_response(request, "no-authtoken")

            # Decode the authtoken cookie.

            authtoken_decoded = self.decode_authtoken(authtoken)
            if not authtoken_decoded:
                return self._create_unauthenticated_response(request, "invalid-authtoken")

            # Sanity check the decoded authtoken.

            if authtoken_decoded["authenticated"] is not True:
                return self._create_unauthenticated_response(request, "invalid-authtoken-auth", authtoken_decoded)

            if self._auth0_client_id != authtoken_decoded[JWT_AUDIENCE_PROPERTY_NAME]:
                return self._create_unauthenticated_response(request, "invalid-authtoken-aud", authtoken_decoded)

            domain = self._get_domain(request)
            if domain != authtoken_decoded["domain"]:
                return self._create_unauthenticated_response(request, "invalid-authtoken-domain", authtoken_decoded)

            # Check the authtoken expiration time (its expiration time must be in the future i.e. greater than now).

            authtoken_expires_time_t = authtoken_decoded["authenticated_until"]
            current_time_t = int(time.time())
            if authtoken_expires_time_t <= current_time_t:
                return self._create_unauthenticated_response(request, "authtoken-expired", authtoken_decoded)

            # If given an env, then check that the specified environment is allowed, i.e. that
            # the request is AUTHORIZED. If not, return HTTP 403 (not authorized); we do NOT,
            # in this case, return 401 (not authenticated), as (above) if not authenticated;
            # the UI acts differently for these two cases. These HTTP status are actually
            # set in via the decorator in react_route_utils.

            if env:
                allowed_envs = authtoken_decoded["allowed_envs"]
                if not self._envs.is_allowed_env(env, allowed_envs):
                    status = "not-authorized-env" if self._envs.is_known_env(env) else "not-authorized-unknown-env"
                    return self._create_unauthorized_response(request, status, authtoken_decoded, but_is_authenticated=True)

            return {**authtoken_decoded, "authorized": True}

        except Exception as e:
            logger.error(f"Authorization exception: {e}")
            return self._create_unauthenticated_response(request, "exception: " + str(e))

    def create_authtoken(self, jwt: str, jwt_expires_at: int, env: str, domain: str) -> str:
        """
        Creates and returns a new signed JWT, to be used as the login authtoken (cookie), from
        the given AUTHENTICATED and signed and encoded JWT, which will contain the following:
        - Booleans indicating authenticated AND authorized.
        - The user name (from the "email" field of the given JWT).
        - The list of known environments.
        - The default environment.
        - The initial environment (i.e. the environment in which the user was first authenticated).
        - The list of allowed (authorized) environments for the user associated with the given JWT.
        - The first/last name of the user associated with the given JWT.
        - The timestamp of when the given JWT was issued (UTC time_t/epoch format).
        - The timestamp of when the given JWT expires issued (UTC time_t/epoch format).
        - The audience (aka "aud" aka Auth0 client ID).
        The allowed environments and first/last name are obtained via the users ElasticSearch store;
        the first/last names are just for informational/display purposes in the client.
        Returns the JWT-signed-encoded authtoken value as a string.
        """
        jwt_decoded = jwt_decode(jwt, self._auth0_client_id, self._auth0_secret)
        email = jwt_decoded.get("email")
        allowed_envs, first_name, last_name = self._envs.get_user_auth_info(email)
        authtoken_decoded = {
            "authenticated": True,
            "authenticated_at": jwt_decoded.get("iat"),
            "authenticated_until": jwt_expires_at,
            "user": email,
            "user_verified": jwt_decoded.get("email_verified"),
            "first_name": first_name,
            "last_name": last_name,
            "allowed_envs": allowed_envs,
            "known_envs": self._envs.get_known_envs(),
            "default_env": self._envs.get_default_env(),
            "initial_env": env,
            "domain": domain
        }
        # JWT-sign-encode the authtoken using our Auth0 client ID (aka audience aka "aud") and
        # secret. This *required* audience is added to the JWT before encoding (done in the
        # jwt_encode function), set to the value we pass here, namely, self._auth0_client_id;
        # it was also in the given JWT (i.e. jwt_decoded["aud"]), and these should match (no
        # need to check); it came from the original Auth0 invocation on the client-side (in the
        # Auth0Lock box); we communicate it to the client-side via the non-protected /header endpoint.
        return jwt_encode(authtoken_decoded, audience=self._auth0_client_id, secret=self._auth0_secret)

    def decode_authtoken(self, authtoken: str) -> dict:
        """
        Fully verifies AND decodes and returns the given JWT-signed-encoded authtoken (cookie).
        If not verified (by the jwt_decode function) then None will be returned.
        See create_authtoken (above) for an enumeration of the contents of the authtoken.
        Returns the verified/decoded JWT as a dictionary.
        """
        return jwt_decode(authtoken, self._auth0_client_id, self._auth0_secret)

    def _create_unauthorized_response(self, request: dict, status: str,
                                      authtoken_decoded: dict, but_is_authenticated: bool) -> dict:
        """
        Creates a response suitable for a request which is NOT authorized, or NOT authenticated,
        depending on authenticated argument. Note that we still want to return some basic info,
        i.e. list of known environments, default environment, domain, and aud (Auth0 client ID)
        is required for the Auth0 login box (Auth0Lock) on the client-side (i.e. React UI). This
        info is gotten from the given decoded authtoken or if not set then sets this info explicitly.
        """
        if authtoken_decoded:
            response = authtoken_decoded
        else:
            response = {
                # 2022-10-18
                # No longer including known-envs in unauthorized responses.
                # "known_envs": self._envs.get_known_envs(),
                "default_env": self._envs.get_default_env(),
                "domain": self._get_domain(request),
                JWT_AUDIENCE_PROPERTY_NAME: self._auth0_client_id  # Needed for Auth0Lock login box on client-side.
            }
        response["authenticated"] = but_is_authenticated
        response["authorized"] = False
        response["status"] = status
        return response

    def _create_unauthenticated_response(self, request: dict, status: str, authtoken_decoded: dict = None) -> dict:
        """
        Creates a response suitable for a request which is NOT authenticated.
        """
        return self._create_unauthorized_response(request, status, authtoken_decoded, but_is_authenticated=False)

    @staticmethod
    def _get_domain(request: dict) -> str:
        return request.get("headers", {}).get("host")

    def get_aws_credentials(self, env: str) -> dict:
        """
        Returns basic AWS credentials info (NOT the secret).
        This is just for informational/display purposes in the UI.
        This has nothing to do with the rest of the authentication
        and authorization stuff here but vaguely related so here seems fine.
        """
        aws_credentials = Auth._cached_aws_credentials.get(env)
        if not aws_credentials:
            try:
                session = boto3.session.Session()
                credentials = session.get_credentials()
                access_key_id = credentials.access_key
                region_name = session.region_name
                caller_identity = boto3.client("sts").get_caller_identity()
                user_arn = caller_identity["Arn"]
                account_number = caller_identity["Account"]
                aws_credentials = {
                    "aws_account_number": account_number,
                    "aws_user_arn": user_arn,
                    "aws_access_key_id": access_key_id,
                    "aws_region": region_name,
                    "auth0_client_id": self._auth0_client_id
                }
                # Try getting the account name though probably no permission at the moment.
                aws_account_name = None
                try:
                    aws_credentials["aws_account_name"] = (
                        boto3.client('iam').list_account_aliases()['AccountAliases'][0])
                except Exception as e:
                    logger.warning(f"Exception (not fatal) getting AWS account alias: {e}")
                if not aws_account_name:
                    try:
                        aws_credentials["aws_account_name"] = (
                            boto3.client('organizations').
                            describe_account(AccountId=account_number).get('Account').get('Name')
                        )
                    except Exception as e:
                        logger.warning(f"Exception (not fatal) getting AWS account name: {e}")
                Auth._cached_aws_credentials[env] = aws_credentials
            except Exception as e:
                logger.warning(f"Exception (not fatal) getting AWS account info: {e}")
                return {}
        return aws_credentials

    def cache_clear(self) -> None:
        self._cached_aws_credentials = {}



from chalice import Response, __version__ as chalice_version
import copy
import datetime
import json
import os
import pkg_resources
import platform
import requests
import socket
import time
from typing import Optional, Union
import urllib.parse
from itertools import chain
from dcicutils.env_utils import EnvUtils, get_foursight_bucket, get_foursight_bucket_prefix, full_env_name
from dcicutils import ff_utils
from dcicutils.misc_utils import get_error_message
from dcicutils.obfuscation_utils import obfuscate_dict
from dcicutils.secrets_utils import get_identity_name, get_identity_secrets
from ...app import app
from ...decorators import Decorators
from ...route_prefixes import ROUTE_PREFIX
from .auth import Auth
from .auth0_config import Auth0Config
from .aws_s3 import AwsS3
from .checks import Checks
from .cookie_utils import create_delete_cookie_string, create_set_cookie_string, read_cookie
from .datetime_utils import convert_datetime_to_time_t, convert_utc_datetime_to_useastern_datetime_string
from .encoding_utils import base64_decode_to_json
from .envs import Envs
from .gac import Gac
from .misc_utils import (
    get_request_arg,
    is_running_locally,
    sort_dictionary_by_case_insensitive_keys
)
from .react_routes import ReactRoutes
from .react_ui import ReactUi


class ReactApi(ReactRoutes):

    CONTENT_TYPE = "Content-Type"
    JSON_CONTENT_TYPE = "application/json"
    STANDARD_HEADERS = {CONTENT_TYPE: JSON_CONTENT_TYPE}

    _cached_header = {}

    def __init__(self):
        super(ReactApi, self).__init__()
        self._envs = Envs(app.core.get_unique_annotated_environment_names())
        self._auth0_config = Auth0Config(app.core.get_portal_url(self._envs.get_default_env()))
        self._auth = Auth(self._auth0_config.get_client(), self._auth0_config.get_secret(), self._envs)
        self._checks = Checks(app.core.check_handler.CHECK_SETUP, self._envs)
        self._react_ui = ReactUi(self)

    @staticmethod
    def create_response(http_status: int = 200,
                        body: Union[dict, list] = None,
                        headers: dict = None,
                        content_type: str = JSON_CONTENT_TYPE) -> Response:
        if not body:
            body = {}
        if not headers:
            headers = ReactApi.STANDARD_HEADERS
        if content_type:
            if id(headers) == id(ReactApi.STANDARD_HEADERS):
                headers = {**headers}
            headers[ReactApi.CONTENT_TYPE] = content_type
        return Response(status_code=http_status, body=body, headers=headers)

    @staticmethod
    def create_success_response(body: Union[dict, list] = None, content_type: str = JSON_CONTENT_TYPE) -> Response:
        if not body:
            body = {}
        return ReactApi.create_response(http_status=200, body=body, content_type=content_type)

    @staticmethod
    def create_redirect_response(location: str, body: dict = None, headers: dict = None) -> Response:
        if not body:
            body = {}
        if not headers:
            if not location:
                raise Exception("No location specified for HTTP redirect.")
            headers = {"Location": location}
        elif location:
            headers = copy.deepcopy(headers)
            headers["Location"] = location
        elif "location" not in [key.lower() for key in headers.keys()]:
            raise Exception("No location specified in header for HTTP redirect.")
        return ReactApi.create_response(http_status=302, body=body, headers=headers)

    @staticmethod
    def create_not_implemented_response(request: dict) -> Response:
        method = request.get("method")
        context = request.get("context")
        path = context.get("path") if isinstance(context, dict) else None
        body = {"error": "Not implemented.", "method": method, "path": path}
        return ReactApi.create_response(http_status=501, body=body)

    @staticmethod
    def create_forbidden_response() -> Response:
        """
        Note that this is different from the unauthenticated and/or unauthorized response
        if the user is not logged in or does not have access to the given environment.
        This is for other forbidden cases, e.g. access to static files we restrict
        access to, or a failed login/authentication attempt.
        """
        return ReactApi.create_response(http_status=403, body={"status": "Forbidden."})

    @staticmethod
    def create_error_response(message: Union[str, Exception]) -> Response:
        if isinstance(message, Exception):
            message = get_error_message(message)
        return ReactApi.create_response(http_status=500, body={"error": message})

    def _get_redirect_url(self, request: dict, env: str, domain: str, context: str) -> str:
        """
        Returns the redirect URL to the UI from the reactredir cookie,
        or if that is not set then to the /login page of the UI.
        """
        redirect_url = read_cookie(request, "reactredir")
        if not redirect_url:
            if is_running_locally(request):
                scheme = "http"
                # Using ROUTE_PREFIX here instead of context because may be just "/" for the local
                # deploy case because we get here via the /callback endpoint (during authentication).
                # Due to confusion between local deploy not implicitly using /api as context so setting
                # it explicitly; this is just so we are dealing with the same paths for either case.
                context = ROUTE_PREFIX
            else:
                scheme = "https"
            if not env:
                env = self._envs.get_default_env()
            if not context:
                context = "/"
            elif not context.endswith("/"):
                context = context + "/"
            redirect_url = f"{scheme}://{domain}{context}react/{env}/login"
        else:
            # Not certain if by design but the React library (universal-cookie) used to
            # write cookies URL-encodes them; rolling with it for now and URL-decoding here.
            redirect_url = urllib.parse.unquote(redirect_url)
        return redirect_url

    def _get_authentication_callback_url(self, request: dict) -> str:
        """
        Returns the URL for our authentication callback endpoint.
        Note this callback endpoint is (still) defined in the legacy Foursight routes.py.
        """
        domain, context = app.core.get_domain_and_context(request)
        headers = request.get("headers", {})
        scheme = headers.get("x-forwarded-proto", "http")
        if is_running_locally(request):
            context = "/"
        return f"{scheme}://{domain}{context}callback/?react"

    def is_react_authentication_callback(self, request: dict) -> bool:
        """
        Returns True iff the given Auth0 authentication/login callback request, i.e. from
        the /callback route which is defined in the main routes.py for both React and non-React
        Auth0 authentication/login, is for a React authentication/login. This is communicated
        via "react" URL parameter in the callback URL, which is setup on the React UI side;
        note this was PREVIOUSLY done there via a "react" string in Auth0 "scope", but changed
        so we can get the Auth0 config (e.g. domain) for the POST to Auth0 using our Auth0Config.
        See: react/src/pages/LoginPage/createAuth0Lock.
        """
        return get_request_arg(request, "react") is not None

    def react_authentication_callback(self, request: dict, env: str) -> Response:
        """
        Called by the main authentication callback function (app_utils.auth0_callback)
        if the above is_react_authentication_callback returns True. Performs the actual
        Auth0 authentication for login via the Auth0 (HTTP POST) API. If successful,
        returns a redirect response (to the UI) with a cookie setting for the login
        authtoken. If unsuccessful, returnes a forbidden (HTTP 403) response.
        """

        auth0_code = get_request_arg(request, "code")
        auth0_domain = self._auth0_config.get_domain()
        auth0_client = self._auth0_config.get_client()
        auth0_secret = self._auth0_config.get_secret()
        if not (auth0_code and auth0_domain and auth0_client and auth0_secret):
            return self.create_forbidden_response()

        # Not actually sure what this auth0_redirect_uri is needed
        # for, but needed it does seem to be, for this Auth0 POST;
        # it evidently needs to be (this) authentication callback URL.
        auth0_redirect_uri = self._get_authentication_callback_url(request)
        auth0_payload = {
            'grant_type': 'authorization_code',
            'client_id': auth0_client,
            'client_secret': auth0_secret,
            'code': auth0_code,
            'redirect_uri': auth0_redirect_uri
        }
        auth0_post_url = f"https://{auth0_domain}/oauth/token"

        auth0_payload_json = json.dumps(auth0_payload)
        auth0_headers = ReactApi.STANDARD_HEADERS
        auth0_response = requests.post(auth0_post_url, data=auth0_payload_json, headers=auth0_headers)
        auth0_response_json = auth0_response.json()
        jwt = auth0_response_json.get("id_token")

        if not jwt:
            return self.create_forbidden_response()

        jwt_expires_in = auth0_response_json.get("expires_in")
        jwt_expires_at = convert_datetime_to_time_t(datetime.datetime.utcnow() +
                                                    datetime.timedelta(seconds=jwt_expires_in))
        domain, context = app.core.get_domain_and_context(request)
        authtoken = self._auth.create_authtoken(jwt, jwt_expires_at, env, domain)
        authtoken_cookie = create_set_cookie_string(request, name="authtoken",
                                                    value=authtoken,
                                                    domain=domain,
                                                    expires=jwt_expires_at, http_only=False)
        redirect_url = self._get_redirect_url(request, env, domain, context)
        return self.create_redirect_response(location=redirect_url, headers={"Set-Cookie": authtoken_cookie})

    def react_authorize(self, request: dict, env: Optional[str]) -> dict:
        """
        Exposed for call from "route" decorator for endpoint authentication protection.
        """
        return self._auth.authorize(request, env)

    # ----------------------------------------------------------------------------------------------
    # Below are the implementation functions corresponding directly to the routes in react_routes.
    # ----------------------------------------------------------------------------------------------

    def react_serve_static_file(self, env: str, paths: list) -> Response:
        """
        Called from react_routes for endpoint: /react/{environ}/{paths}
        Serves static UI related (JavaScript, CSS, HTML) files.
        Note that this in an UNPROTECTED route.
        """
        return self._react_ui.serve_static_file(env, paths)

    def reactapi_auth0_config(self, request: dict):
        """
        Called from react_routes for endpoint: /reactapi/{environ}/auth0_config
        Note that this in an UNPROTECTED route.
        """
        auth0_config = self._auth0_config.get_config_data()
        # Note we add the callback for the UI to setup its Auth0 login for.
        auth0_config["callback"] = self._get_authentication_callback_url(request)
        return self.create_success_response(body=self._auth0_config.get_config_data())

    def reactapi_logout(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/logout
        Note that this in an UNPROTECTED route.
        """
        authorize_response = self.react_authorize(request, env)
        if not authorize_response or not authorize_response["authorized"]:
            body = {"status": "Already logged out."}
        else:
            body = {"status": "Logged out."}
        domain, context = app.core.get_domain_and_context(request)
        authtoken_cookie_deletion = create_delete_cookie_string(request=request, name="authtoken", domain=domain)
        redirect_url = self._get_redirect_url(request, env, domain, context)
        headers = {"Set-Cookie": authtoken_cookie_deletion}
        return self.create_redirect_response(location=redirect_url, body=body, headers=headers)

    def reactapi_header(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/header
        Note that this in an UNPROTECTED route.
        """
        # Note that this route is not protected but/and we return the results from authorize.
        auth = self._auth.authorize(request, env)
        data = ReactApi._cached_header.get(env)
        if not data:
            data = self._reactapi_header_nocache(request, env)
            ReactApi._cached_header[env] = data
        data = copy.deepcopy(data)
        data["auth"] = auth
        # 2022-10-18
        # No longer sharing known-envs widely; send only if authenticated;
        # if not authenticated then act as-if the default-env is the only known-env,
        # and in this case also include (as an FYI for the UI) the real number of known-envs.
        if auth["authenticated"]:
            data["auth"]["known_envs"] = self._envs.get_known_envs_with_gac_names()
        else:
            known_envs_default = self._envs.find_known_env(self._envs.get_default_env())
            known_envs_actual_count = self._envs.get_known_envs_count()
            data["auth"]["known_envs"] = [known_envs_default]
            data["auth"]["known_envs_actual_count"] = known_envs_actual_count
        data["timestamp"] = convert_utc_datetime_to_useastern_datetime_string(datetime.datetime.utcnow())
        return self.create_success_response(body=data)

    def _reactapi_header_nocache(self, request: dict, env: str) -> dict:
        """
        No-cache version of above reactapi_header function.
        """
        domain, context = app.core.get_domain_and_context(request)
        stage_name = app.core.stage.get_stage()
        default_env = self._envs.get_default_env()
        aws_credentials = self._auth.get_aws_credentials(env if env else default_env)
        response = {
            "app": {
                "title": app.core.html_main_title,
                "package": app.core.APP_PACKAGE_NAME,
                "stage": stage_name,
                "version": app.core.get_app_version(),
                "domain": domain,
                "context": context,
                "local": is_running_locally(request),
                "credentials": {
                    "aws_account_number": aws_credentials["aws_account_number"],
                    "aws_account_name": aws_credentials["aws_account_name"]
                },
                "launched": app.core.init_load_time,
                "deployed": app.core.get_lambda_last_modified()
            },
            "versions": {
                "foursight": app.core.get_app_version(),
                "foursight_core": pkg_resources.get_distribution('foursight-core').version,
                "dcicutils": pkg_resources.get_distribution('dcicutils').version,
                "python": platform.python_version(),
                "chalice": chalice_version
            },
        }
        return response

    def reactapi_info(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/info
        """
        domain, context = app.core.get_domain_and_context(request)
        stage_name = app.core.stage.get_stage()
        default_env = self._envs.get_default_env()
        if not self._envs.is_known_env(env):
            env_unknown = True
        else:
            env_unknown = False
        if not env_unknown:
            try:
                environment_and_bucket_info = sort_dictionary_by_case_insensitive_keys(
                    obfuscate_dict(app.core.environment.get_environment_and_bucket_info(env, stage_name))),
                portal_url = app.core.get_portal_url(env)
            except Exception as e:
                environment_and_bucket_info = None
                portal_url = None
        else:
            environment_and_bucket_info = None
            portal_url = None
        # Get known envs with GAC name for each.
        body = {
            "app": {
                "title": app.core.html_main_title,
                "package": app.core.APP_PACKAGE_NAME,
                "stage": stage_name,
                "version": app.core.get_app_version(),
                "domain": domain,
                "context": context,
                "local": is_running_locally(request),
                "credentials": self._auth.get_aws_credentials(env if env else default_env),
                "launched": app.core.init_load_time,
                "deployed": app.core.get_lambda_last_modified()
            },
            "versions": {
                "foursight": app.core.get_app_version(),
                "foursight_core": pkg_resources.get_distribution('foursight-core').version,
                "dcicutils": pkg_resources.get_distribution('dcicutils').version,
                "python": platform.python_version(),
                "chalice": chalice_version
            },
            "server": {
                "foursight": socket.gethostname(),
                "portal": portal_url,
                "es": app.core.host,
                "rds": os.environ["RDS_HOSTNAME"],
                # TODO: cache this (slow).
                "sqs": app.core.sqs.get_sqs_queue().url,
            },
            # TODO: cache this (slow).
            "buckets": {
                "env": app.core.environment.get_env_bucket_name(),
                "foursight": get_foursight_bucket(envname=env if env else default_env, stage=stage_name),
                "foursight_prefix": get_foursight_bucket_prefix(),
                "info": environment_and_bucket_info,
                "ecosystem": sort_dictionary_by_case_insensitive_keys(EnvUtils.declared_data()),
            },
            "page": {
                "path": request.get("context").get("path"),
                "endpoint": request.get("path"),
                "loaded": app.core.get_load_time()
            },
            # TODD: Move these out to another API.
            "checks": {
                "running": 0,
                "queued": 0
            },
            "known_envs": self._envs.get_known_envs_with_gac_names(),
            # TODO: cache this (slow).
            "gac": {
                "name": get_identity_name(),
                "values": sort_dictionary_by_case_insensitive_keys(obfuscate_dict(get_identity_secrets())),
             },
            "environ": sort_dictionary_by_case_insensitive_keys(obfuscate_dict(dict(os.environ)))
        }
        return self.create_success_response(body=body)

    def reactapi_users(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/users
        Returns info on all users. TODO: No paging supported yet!
        """
        users = []
        # TODO: Support paging.
        # TODO: Consider adding ability to search for both normal users and
        #       admin/foursight users (who would have access to foursight);
        #       and more advanced, the ability to grant foursight access.
        user_records = ff_utils.get_metadata('users/', ff_env=full_env_name(env), add_on='frame=object&limit=10000')
        for user_record in user_records["@graph"]:
            last_modified = user_record.get("last_modified")
            if last_modified:
                last_modified = last_modified.get("date_modified")
            # TODO
            # roles = []
            # project_roles = user_record.get("project_roles")
            # if project_roles:
            #     role = role.get("date_modified")
            #     roles.append({
            #         "groups": groups,
            #         "project_roles": project_roles,
            #         "principals_view": principals_view,
            #         "principals_edit": principals_edit
            #     })
            users.append({
                "email_address": user_record.get("email"),
                "first_name": user_record.get("first_name"),
                "last_name": user_record.get("last_name"),
                "uuid": user_record.get("uuid"),
                "modified": convert_utc_datetime_to_useastern_datetime_string(last_modified)})
        return self.create_success_response(body=sorted(users, key=lambda key: key["email_address"]))

    def reactapi_get_user(self, request: dict, env: str, email: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/user/{email}
        Returns info on the specified user (email).
        """
        users = []
        for email_address in email.split(","):
            try:
                user = ff_utils.get_metadata('users/' + email_address.lower(),
                                             ff_env=full_env_name(env), add_on='frame=object')
                users.append({"email_address": email_address, "record": user})
            except Exception as e:
                users.append({"email_address": email_address, "record": {"error": str(e)}})
        return self.create_success_response(body=sorted(users, key=lambda key: key["email_address"]))

    def reactapi_checks(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks
        Returns a summary (list) of all defined checks.
        """
        return self.create_success_response(body=self._checks.get_checks_grouped(env))

    def reactapi_check_results(self, request: dict, env: str, check: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks/{check}
        Returns the latest result from the given check (name).
        """
        connection = app.core.init_connection(env)
        check_results = app.core.CheckResult(connection, check)
        check_results = check_results.get_latest_result()
        uuid = check_results["uuid"]
        check_datetime = datetime.datetime.strptime(uuid, "%Y-%m-%dT%H:%M:%S.%f")
        check_datetime = convert_utc_datetime_to_useastern_datetime_string(check_datetime)
        check_results["timestamp"] = check_datetime
        return self.create_success_response(body=check_results)

    def reactapi_check_result(self, request: dict, env: str, check: str, uuid: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks/{check}/{uuid}
        Returns the check result for the given check (name) and uuid.
        Analogous legacy function is app_utils.view_foursight_check.
        """
        body = []
        try:
            connection = app.core.init_connection(env)
        except Exception as e:
            connection = None
        if connection:
            check_result = app.core.CheckResult(connection, check)
            if check_result:
                data = check_result.get_result_by_uuid(uuid)
                if data is None:
                    # the check hasn't run. Return a placeholder view
                    data = {
                        'name': check,
                        'uuid': uuid,
                        'status': 'ERROR',  # in this case we just queued a check, so ERROR is ok
                        'summary': 'Check has not yet run',
                        'description': 'Check has not yet run'
                    }
                title = app.core.check_handler.get_check_title_from_setup(check)
                processed_result = app.core.process_view_result(connection, data, is_admin=True)
                body.append({
                    'status': 'success',
                    'env': env,
                    'checks': {title: processed_result}
                })
        return self.create_success_response(body=body)

    def reactapi_checks_history(self, request: dict, env: str,
                                check: str, offset: int = 0, limit: int = 25, sort: str = None) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks/check/history
        Returns a (paged) summary (list) of check results for the given check (name).
        """
        if offset < 0:
            offset = 0
        if limit < 0:
            limit = 0
        check_record = self._checks.get_check(env, check)
        connection = app.core.init_connection(env)
        history, total = app.core.get_foursight_history(connection, check, offset, limit, sort)
        history_kwargs = list(set(chain.from_iterable([item[2] for item in history])))
        queue_attr = app.core.sqs.get_sqs_attributes(app.core.sqs.get_sqs_queue().url)
        for item in history:
            for subitem in item:
                if isinstance(subitem, dict):
                    uuid = subitem.get("uuid")
                    if uuid:
                        timestamp = datetime.datetime.strptime(uuid, "%Y-%m-%dT%H:%M:%S.%f")
                        timestamp = convert_utc_datetime_to_useastern_datetime_string(timestamp)
                        subitem["timestamp"] = timestamp
        body = {
            "check": check_record,
            "env": env,
            "history_kwargs": history_kwargs,
            "paging": {
                "total": total,
                "count": len(history),
                "limit": min(limit, total),
                "offset": min(offset, total),
                "more": max(total - offset - limit, 0)
            },
            "list": history
        }
        return self.create_success_response(body=body)

    def reactapi_checks_run(self, request: dict, env: str, check: str, args: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks/{check}/run
        Kicks off a run for the given check (name).
        """
        args = base64_decode_to_json(args)
        queued_uuid = app.core.queue_check(env, check, args)
        return self.create_success_response(body={"check": check, "env": env, "uuid": queued_uuid})

    def reactapi_checks_status(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks-status
        Returns the status of any/all currently running or queued checks.
        """
        checks_queue = app.core.sqs.get_sqs_attributes(app.core.sqs.get_sqs_queue().url)
        checks_running = checks_queue.get('ApproximateNumberOfMessagesNotVisible')
        checks_queued = checks_queue.get('ApproximateNumberOfMessages')
        return self.create_success_response(body={"checks_running": checks_running, "checks_queued": checks_queued})

    def reactapi_checks_raw(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks-raw
        Returns the content of the raw/original check_setup.json file.
        """
        return self.create_success_response(body=self._checks.get_checks_raw())

    def reactapi_checks_registry(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/checks-registry
        Returns the content of the checks registry collected for the check_function
        decorator in decorators.py.
        """
        return self.create_success_response(body=Decorators.get_registry())

    def reactapi_lambdas(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/lambdas
        Returns a summary (list) of all defined AWS lambdas for the current AWS environment.
        """
        return self.create_success_response(body=self._checks.get_annotated_lambdas())

    def reactapi_gac_compare(self, request: dict, env: str, env_compare: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/gac/{environ_compare}
        Returns differences between two GACs (global application configurations).
        """
        return self.create_success_response(body=Gac.compare_gacs(env, env_compare))

    def reactapi_aws_s3_buckets(self, request: dict, env: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/s3/buckets
        Return a list of all AWS S3 bucket names for the current AWS environment.
        """
        return self.create_success_response(body=AwsS3.get_buckets())

    def reactapi_aws_s3_buckets_keys(self, request: dict, env: str, bucket: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/s3/buckets/{bucket}
        Return a list of all AWS S3 bucket key names in the given bucket
        for the current AWS environment.
        """
        return self.create_success_response(body=AwsS3.get_bucket_keys(bucket))

    def reactapi_aws_s3_buckets_key_contents(self, request: dict, env: str, bucket: str, key: str) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/{environ}/s3/buckets/{bucket}/{key}
        Return the contents of the AWS S3 bucket key in the given bucket for the current AWS environment.
        """
        if True:
            #
            # TODO!!!
            # Disabling this feature for now until we can discuss/resolve security concerns.
            #
            return self.create_not_implemented_response(request)
        key = urllib.parse.unquote(key)
        return self.create_success_response(body=AwsS3.get_bucket_key_contents(bucket, key))

    def reactapi_reload_lambda(self, request: dict) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/__reloadlambda__
        Kicks off a reload of the given lambda name. For troubleshooting only.
        """
        app.core.reload_lambda()
        time.sleep(3)
        return self.create_success_response(body={"status": "OK"})

    def reactapi_clear_cache(self, request: dict) -> Response:
        """
        Called from react_routes for endpoint: /reactapi/__clearcache___
        Not yet implemented.
        """
        self.cache_clear()
        return self.create_success_response(body={"status": "OK"})

    def cache_clear(self) -> None:
        self._auth.cache_clear()
        self._auth0_config.cache_clear()
        self._envs.cache_clear()
        self._checks.cache_clear()
        Gac.cache_clear()
        self._cached_header = {}



# This module defines a "route" decorator which wraps the Chalice route decorator,
# and does authorization (and authentication) checking, tweaks the path appropriately,
# sets up CORS if necessary (only for local development), and has common exception handling.
# We DEFAULT to AUTHORIZATION checking; if not wanted use authorize=False in route decorator.

from chalice import CORSConfig, Response
from typing import Optional, Tuple
from ...app import app
from ...route_prefixes import ROUTE_CHALICE_LOCAL, ROUTE_PREFIX

# CORS is need ONLY for local DEVELOPMENT, i.e. when using chalice local!
# Set CORS to True if CHALICE_LOCAL; not needed if running React from Foursight
# directly, on the same port (e.g. 8000), but useful if/when running React on a
# separate port (e.g. 3000) via npm start in foursight-core/react to facilitate
# easy/quick development/changes directly to React UI code.
if ROUTE_CHALICE_LOCAL:
    # Very specific/tricky requirements for running Foursight React UI/API in CORS
    # mode (i.e. UI on localhost:3000 and API on localhost:8000). The allow_origin
    # must be exact (i.e. no "*" allowed), and the allow_credentials must be True.
    # On the client-side (React UI) we must include 'credentials: "include"' in the
    # fetch (if using React fetch, though now we use axios, where we require, in
    # any case, 'withCredentials: "include"').
    _CORS = CORSConfig(
        allow_origin="http://localhost:3000",  # Need this to be explicit not "*"
        allow_credentials=True  # Need this
    )
else:
    _CORS = None

_HTTP_UNAUTHENTICATED = 401
_HTTP_UNAUTHORIZED = 403


def route(*args, **kwargs):
    """
    Decorator to wrap the Chalice route decorator to do our authentication and
    authorization checking; also tweaks the path appropriately; sets up CORS
    if necessary (only for local development); and handles exceptions.
    """
    if not isinstance(args, Tuple) or len(args) == 0:
        raise Exception("No arguments found for route configuration!")

    path = args[0]
    route_prefix = ROUTE_PREFIX + ("/" if not ROUTE_PREFIX.endswith("/") else "")
    if "static" in kwargs:
        # This is for serving static files which live in a different/specific directory.
        route_prefix = route_prefix + "react"
        del kwargs["static"]
    else:
        route_prefix = route_prefix + "reactapi"
    path = route_prefix + path
    if path.endswith("/"):
        path = path[:-1]

    if "authorize" in kwargs:
        authorize = kwargs["authorize"]
        authorize = isinstance(authorize, bool) and authorize
        del kwargs["authorize"]
    else:
        # Note we DEFAULT to AUTHORIZE!
        # Only way to turn it off is to explicitly pass authorize=False to the route decorator.
        authorize = True

    def route_registration(wrapped_route_function):
        """
        This function is called once for each defined route/endpoint (at app startup).
        """
        def route_function(*args, **kwargs):
            """
            This is the function called on each route/endpoint (API) call.
            """
            try:
                if authorize:
                    # Was calling the _route_requires_authorization decorator; now do it directly.
                    # return _route_requires_authorization(wrapped_route_function)(*args, **kwargs)
                    # Note that we access the env argument in kwargs; ASSUME this name from the route.
                    unauthorized_response = _authorize(app.current_request.to_dict(), kwargs.get("env"))
                    if unauthorized_response:
                        return unauthorized_response
                return wrapped_route_function(*args, **kwargs)
            except Exception as e:
                # Common endpoint exception handling here.
                return app.core.create_error_response(e)
        if _CORS:
            # Only used for cross-origin localhost development (e.g. UI on 3000 and API on 8000).
            kwargs["cors"] = _CORS
        # This is the call that registers the Chalice route/endpoint.
        app.route(path, **kwargs)(route_function)
        return route_function
    return route_registration


def _authorize(request: dict, env: Optional[str]) -> Optional[Response]:
    """
    If the given request is unauthorized (or unauthenticated) then return
    an appropriate unauthorized (or unauthenticated) response, otherwise None.
    """
    authorize_response = app.core.react_authorize(request, env)
    if not authorize_response or not authorize_response["authorized"]:
        # HTTP 401 - Unauthorized (more precisely: Unauthenticated):
        # Request has no or invalid credentials.
        # HTTP 403 - Forbidden (more precisely: Unauthorized):
        # Request has valid credentials but no privileges for resource.
        http_status = _HTTP_UNAUTHENTICATED if not authorize_response["authenticated"] else _HTTP_UNAUTHORIZED
        return app.core.create_response(http_status=http_status, body=authorize_response)
    return None


def _route_requires_authorization(f):
    """
    THIS DECORATOR NOT USED ANYMORE.
    Functionality now wrapped in common route decorator above.

    Decorator for Chalice routes which should be PROTECTED by an AUTHORIZATION check.
    This ASSUMES that the FIRST argument to the route function using this decorator
    is the ENVIRONMENT name. The Chalice request is gotten from app.current_request.

    If the request is NOT authorized/authenticated then a forbidden (HTTP 403 or 401)
    response is returned, otherwise we go ahead with the function/route invocation.
    A request is authorized iff it is AUTHENTICATED, i.e. the user has successfully
    logged in, AND also has PERMISSION to access the specified environment. If the
    request is not authorized an HTTP 401 is returned; if the request is authenticated
    but is not authorized (for the specified environment) and HTTP 403 is returned.

    The info to determine this is pass via the authtoken cookie, which is a (server-side)
    JWT-signed-encode value containing authentication info and list allowed environment
    for the user; this value/cookie is set server-side at login time.

    Note that ONLY three React API routes should NOT be authorization protected by this decorator:
      -> /{env}/auth0_config
      -> /{env}/header
      -> /{env}/logout

    N.B. This was originally used to decorate each route, then created the below "route"
    wrapper which called this. And then changed that (below) "route" wrapper to call
    the function directly (maybe more efficient); so this is actually NO LONGER USED.
    Keeping here just for now just in case.

    """
    def wrapper(*args, **kwargs):
        if not kwargs or len(kwargs) < 1:
            raise Exception("Invalid arguments to requires_authorization decorator!")
        unauthorized_response = _authorize(app.current_request.to_dict(), kwargs.get("env"))
        if unauthorized_response:
            return unauthorized_response
        return f(*args, **kwargs)
    return wrapper



from typing import Optional

def sort_dictionary_by_case_insensitive_keys(dictionary: dict) -> dict:
    """
    Returns the given dictionary sorted by (case-insenstivie) key values; yes,
    dictionaries are ordered as of Python 3.7. If the given value is not a
    dictionary it will be coerced to one.
    :param dictionary: Dictionary to sort.
    :return: Given dictionary sorted by key value.
    """
    if not dictionary or not isinstance(dictionary, dict):
        return {}
    return {key: dictionary[key] for key in sorted(dictionary.keys(), key=lambda key: key.lower())}


def get_request_arg(request: dict, name: str) -> Optional[str]:
    """
    Returns the value of the given URL query parameter name for the given request;
    returns None if not present. Note if present but with no value, then and empty
    string is returned, e.g. for http://localhost:8000/callback?react if "react"
    is queried then returns an empty string, and not None.
    """
    query_params = request.get("query_params")
    return query_params.get(name, None) if query_params else None


def is_running_locally(request: dict) -> bool:
    """
    Returns true iff the given request indicates that we are running locally (localhost).
    """
    return request.get("context", {}).get("identity", {}).get("sourceIp", "") == "127.0.0.1"



from chalice import Response
import io
import logging
import os
from typing import Optional

logging.basicConfig()
logger = logging.getLogger(__name__)

_REACT_BASE_DIR = os.path.join(os.path.dirname(__file__), "../../react/ui")
_REACT_DEFAULT_FILE = "index.html"
_REACT_STATIC_FILE_TYPES = [
    {"suffix":       ".html",
     "content_type": "text/html",
     "open_mode":    "r"},
    {"suffix":       ".js",
     "content_type": "text/javascript",
     "open_mode":    "r"},
    {"suffix":       ".css",
     "content_type": "application/css",
     "open_mode":    "r"},
    {"suffix":       ".json",
     "content_type": "application/json",
     "open_mode":    "r"},
    {"suffix":       ".png",
     "content_type": "image/png",
     "open_mode":    "rb"},
    {"suffix":       ".jpg",
     "content_type": "image/jpeg",
     "open_mode":    "rb"},
    {"suffix":       ".jpeg",
     "content_type": "image/jpeg",
     "open_mode":    "rb"},
    {"suffix":       ".ico",
     "content_type": "image/x-icon",
     "open_mode":    "rb"}
]
_REACT_WHITELISTED_FILE_PATH_SUFFIXES = [
    "/index.html",
    "/main.js",
    "/main.css",
    "/manifest.json",
    "/asset-manifest.json",
    ".jpeg",
    ".jpg",
    ".png",
    ".ico"
]


class ReactUi:

    def __init__(self, react_api):
        self._react_api = react_api

    _cache_static_files = {}

    @staticmethod
    def _is_known_file_suffix(file: str) -> bool:
        return any(file.endswith(file_info["suffix"]) for file_info in _REACT_STATIC_FILE_TYPES)

    @staticmethod
    def _get_file_info(file: str) -> Optional[dict]:
        for info in _REACT_STATIC_FILE_TYPES:
            if file.endswith(info["suffix"]):
                return info
        return None

    @staticmethod
    def _is_file_whitelisted(file: str) -> bool:
        """
        To be as restrictive as possible we ONLY allow the above whitelistted files.
        """
        for suffix in _REACT_WHITELISTED_FILE_PATH_SUFFIXES:
            if file.endswith(suffix):
                return True
        return False

    def serve_static_file(self, env: str, paths: list) -> Response:

        if env == "static":
            # If the env is 'static' then we take this to mean the 'static' subdirectory;
            # this is the directory where the static (js, css, etc) React files reside.
            # Note this means that an environment name may not be the literal string 'static'.
            file = os.path.join(_REACT_BASE_DIR, "static")
        else:
            file = _REACT_BASE_DIR
        if not paths:
            # TODO: Not downloading png (et.al.) right! Works with chalice local!
            # Actually it also works in cgap-supertest:
            # https://810xasmho0.execute-api.us-east-1.amazonaws.com/api/react/logo192.png
            # But not in 4dn/foursight-development:
            # https://cm3dqx36s7.execute-api.us-east-1.amazonaws.com/api/react/logo192.png
            # Anyways for now the React UI references images at external sites not from here.
            if self._is_known_file_suffix(env):
                # If the env appears to refer to a file name then we take this
                # to mean a file in the main React directory. Note this means means
                # the environment name may NOT be a value ending in the above suffixes.
                paths = [env]
        for path in paths:
            file = os.path.join(file, path)

        file_info = self._get_file_info(file)
        if file_info:
            content_type = file_info["content_type"]
            open_mode = file_info["open_mode"]
        else:
            # If not recognized then serve the default (index.html) file;
            # this is actually the common case of getting any React UI path.
            file = os.path.join(_REACT_BASE_DIR, _REACT_DEFAULT_FILE)
            content_type = "text/html"
            open_mode = "r"
        file = os.path.normpath(file)

        # Restrict to known whitelisted files.
        if not self._is_file_whitelisted(file):
            return self._react_api.create_forbidden_response()

        response = ReactUi._cache_static_files.get(file)
        if not response:
            response = self._react_api.create_success_response(content_type=content_type)
            try:
                with io.open(file, open_mode) as f:
                    response.body = f.read()
            except Exception as e:
                message = f"Exception serving static React file ({file} | {content_type}): {e}"
                logger.error(message)
                return self._react_api.create_error_response(message)
            ReactUi._cache_static_files[file] = response = self._react_api.process_response(response)
        return response



import datetime
import pytz


EPOCH = datetime.datetime.utcfromtimestamp(0)  # I.e.: 1970-01-01 00:00:00 UTC


def convert_utc_datetime_to_useastern_datetime_string(t: datetime.datetime) -> str:
    """
    Converts the given UTC datetime object or string into a US/Eastern datetime string
    and returns its value in a form that looks like 2022-08-22 13:25:34 EDT.
    If the argument is a string it is ASSUMED to have a value which looks
    like 2022-08-22T14:24:49.000+0000; this is the datetime string format
    we get from AWS via boto3 (e.g. for a lambda last-modified value).

    :param t: UTC datetime object or string value.
    :return: US/Eastern datetime string (e.g.: 2022-08-22 13:25:34 EDT).
    """
    try:
        if isinstance(t, str):
            t = datetime.datetime.strptime(t, "%Y-%m-%dT%H:%M:%S.%f%z")
        t = t.replace(tzinfo=pytz.UTC).astimezone(pytz.timezone("US/Eastern"))
        return t.strftime("%Y-%m-%d %H:%M:%S %Z")
    except:
        return "datetime-utc-parse-error"


def convert_time_t_to_useastern_datetime_string(time_t: int) -> str:
    """
    Converts the given "epoch" time (seconds since 1970-01-01T00:00:00Z)
    integer value to a US/Eastern datetime string and returns its value
    in a form that looks like 2022-08-22 13:25:34 EDT.

    :param time_t: Epoch time value (i.e. seconds since 1970-01-01T00:00:00Z)
    :return: US/Eastern datetime string (e.g.: 2022-08-22 13:25:34 EDT).
    """
    try:
        if not isinstance(time_t, int):
            return ""
        t = datetime.datetime.fromtimestamp(time_t, tz=pytz.UTC)
        return convert_utc_datetime_to_useastern_datetime_string(t)
    except:
        return "datetime-parse-error"

def convert_time_t_to_datetime(time_t: int) -> datetime.datetime:
    """
    Converts the given "epoch" time (seconds since 1970-01-01T00:00:00Z)
    to a datetime and returns its value.
    """
    return datetime.datetime.fromtimestamp(time_t, tz=pytz.UTC)

def convert_datetime_to_time_t(value: datetime.datetime) -> int:
    """
    Converts the given datatime value to an "epoch"
    time (seconds since 1970-01-01T00:00:00Z) and returns its value.
    """
    return int(value.timestamp())

def convert_utc_datetime_to_cookie_expires_format(t) -> str:
    """
    Converts the given UTC datetime object or string format suitable
    for use by a cookie expires date/time value.
    """
    return t.strftime("%a, %d %b %Y %H:%M:%S GMT")
